#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>

/* libusb */
#include <libusb.h>

/* #defines */
#define USB_VID_RAZER 0x1532
#define USB_PID_RAZER_BLADE 0x0116

/* structures, enums, stuff that generally goes into a header */
struct _razer_device {
	/* libusb */
	libusb_device_handle *handle;
	libusb_device *dev;
	struct libusb_config_descriptor *cfg;
	libusb_context *ctx;

	/* Razer */
	// XXX: notch is an array whose length is determined
	//      by razer_device->cfg->bNumInterfaces
	struct _razer_interface *notch;
	uint8_t did_set_altinterface;
	time_t last_input_sent; // NOTE: used to time-out the test
};

typedef struct _razer_device razer_device;

// NOTE: "Controller" here refers to the "device-specific" interface
//       and endpoints most likely used for touchpad/graphics.
//       (Hopefully not firmware... that would blow)
typedef enum {
	RAZER_TRACKPAD   = 0,
	RAZER_KEYBOARD   = 2,
	RAZER_MACROS     = 1,
	RAZER_CONTROLLER = 3,
	RAZER_UNKNOWN    = 4
} razer_input_type;

typedef enum {
	RAZER_MUXED_KEYBOARD = 1,
	RAZER_MUXED_XTRA     = 4,
	RAZER_MUXED_UNKNOWN  = 0
} razer_mux_input_type;

#define RAZER_IS_MACROKEY 0x50
#define RAZER_FN_KEY 0x01
#define RAZER_RAZER_KEY 0x02

// I'm not sure when I will ever use this besides that one function.
typedef enum {
	LIBRAZER_DEV_NOT_INITED  = 1,
	LIBRAZER_DEV_NOT_CORRECT = 2,
	LIBRAZER_COULDNT_SET_ALT = 3,
	LIBRAZER_SUCCESS         = 0
} razer_result;

struct _razer_interface {
	/* libusb */
	struct libusb_transfer *transfer;

	/* Razer */
	razer_device *blade;
	razer_input_type type;

	/* Per-interface */
	int interface_number;
	unsigned char *buffer;
	size_t buffer_size;
	uint8_t kdrv_detached;
	uint8_t altinterface_set;
};

typedef struct _razer_interface razer_interface;

/* Prototypes */
void free_razer_device(razer_device *blade);
razer_device *get_razer_device(libusb_context *ctx, time_t timeout);
void razer_input_callback(struct libusb_transfer *transfer);
razer_result razer_device_initinput(razer_device *blade); 
razer_result razer_device_set_alt(razer_device *blade, int alternate_setting, time_t timeout);

/* code */
razer_device *get_razer_device(libusb_context *ctx, time_t timeout) {
	/* vars */
	razer_device *blade;
	int result;
	uint8_t claimed;

	/* libusb must be initialized before this can be called */
	/* (can't check here for that -- TODO) */
	blade = (razer_device *)malloc(sizeof(razer_device));
	memset(blade, 0, sizeof(razer_device));

	/* So find our Razer device */
	blade->handle = libusb_open_device_with_vid_pid(ctx, USB_VID_RAZER, USB_PID_RAZER_BLADE);

	if (!blade->handle) {
		// No Blade found, free and return NULL
		free(blade);
		return NULL;
	} else {
		printf("Found Razer Blade\n");
	}

	blade->ctx = ctx;
	blade->dev = libusb_get_device(blade->handle);
	libusb_get_active_config_descriptor(blade->dev, &(blade->cfg));

	// Start working at the interfaces
	blade->notch = (razer_interface*)malloc(sizeof(razer_interface) * blade->cfg->bNumInterfaces);
	memset(blade->notch, 0, sizeof(razer_interface) * blade->cfg->bNumInterfaces);

	// Maybe we can cooperatively claim interfaces. Worth a shot right? 
	for (claimed = 0; claimed < blade->cfg->bNumInterfaces; claimed++) {
		// If a kernel driver is attached, detach before claiming.
		result = libusb_kernel_driver_active(blade->handle, claimed);
		if (result) {
			if (result == 1) {
				// Temporarily detach kernel driver.
				result = libusb_detach_kernel_driver(blade->handle, claimed);
				if (result) {
					// couldn't detach the driver
					// bail
					printf("Couldn't detach the kernel driver: BAILING on libusb error %s\n", libusb_error_name(result));
					free_razer_device(blade);
					return NULL;
				}

				blade->notch[claimed].kdrv_detached = 1;
			} else {
				// welp, couldn't see kernel driver status
				// bail
				printf("Couldn't ascertain kernel driver attachment status: BAILING on libusb error %s\n", libusb_error_name(result));
				free_razer_device(blade);
				return NULL;
			}
		}

		result = libusb_claim_interface(blade->handle, claimed);
		if (result) {
			// claim failed. bail
			printf("Claiming the interface failed: BAILING on libusb error %s\n", libusb_error_name(result));
			free_razer_device(blade);
			return NULL;
		}

		// claimed! make it work!
		blade->notch[claimed].blade = blade;
		blade->notch[claimed].type = claimed; // [1]
		blade->notch[claimed].interface_number = claimed;

		/* allocate the transfer request */
		if (blade->notch[claimed].type != RAZER_CONTROLLER) {
			blade->notch[claimed].buffer_size = blade->cfg->interface[claimed].altsetting[0].endpoint->wMaxPacketSize;
			blade->notch[claimed].buffer = (unsigned char*)malloc(sizeof(unsigned char) * blade->notch[claimed].buffer_size);

			blade->notch[claimed].transfer = libusb_alloc_transfer(0);
			if (!blade->notch[claimed].transfer) {
				printf("ERROR: Couldn't allocate transfer request for interface %i; BAILING!\n", claimed);
				free_razer_device(blade);
				return NULL;
			}

			// it's an interrupt type, so set it up as such
			libusb_fill_interrupt_transfer(blade->notch[claimed].transfer,
			                               blade->handle,
						       blade->cfg->interface[claimed].altsetting[0].endpoint->bEndpointAddress,
						       blade->notch[claimed].buffer,
						       blade->notch[claimed].buffer_size,
						       &razer_input_callback,
						       (void*)&(blade->notch[claimed]),
						       timeout
						      );
		} else {
			// Surprisingly, there are no "input" transfers for the controller
		}
	}

	blade->last_input_sent = time(NULL);

	return blade;
}

/* free_razer_device */
void free_razer_device(razer_device *blade) {
	/* Variables */
	int result = 0;
	uint8_t interface = 0;

	// We're in sort of an odd position here, because
	// we could be called from the device initialization process
	// in case of an error -- so this is a general-purpose
	// cleanup function. 

	// First, clean up the device interfaces, if they were created
	printf("Freeing Razer Device...\n");
	if (blade->notch) {
		for (interface = 0; interface < blade->cfg->bNumInterfaces; interface++) {
			if (!blade->notch[interface].blade) {
				// this interface isn't valid -- most likely failed to
				// be claimed. Ignore.
				continue;
			}
	
			if (blade->notch[interface].altinterface_set) {
				// re-set everything back
				// XXX: we have a function to do this now, but it works
				//      across the whole device. Eh.
				//
				// Additional note: this could probably be determined via
				// a separate libusb configuration setting check, but
				// it's somewhat easier to just do it here.
				printf("Resetting altinterface back to 0\n");
			//	result = libusb_set_interface_alt_setting(blade->handle, blade->notch[interface].interface_number, 0);
				result = razer_device_set_alt(blade, 0, 10); // we're about to free it anyway!
				if (result) {
					printf("Couldn't reset altinterface back to 0: libusb error %s\n", libusb_error_name(result));
				}
			}

			/* razer_input_callback now handles freeing for the interfaces */ 
/*
			result = libusb_release_interface(blade->handle, blade->notch[interface].interface_number);
			if (result) {
				printf("free_razer_device: WARNING: Unable to release interface: %s\n", libusb_error_name(result));
			}

			if (blade->notch[interface].kdrv_detached) {
				// Re-attach the kernel driver -- prevent total loss of input
				result = libusb_attach_kernel_driver(blade->handle, blade->notch[interface].interface_number);
				if (result) {
					printf("free_razer_device: WARNING: Couldn't re-attach kernel driver to Blade interface %i: libusb error %s\n", blade->notch[interface].interface_number, libusb_error_name(result));
				}
			}

			if (blade->notch[interface].buffer) {
				free(blade->notch[interface].buffer);
			}

			if (blade->notch[interface].transfer) {
				libusb_free_transfer(blade->notch[interface].transfer);
			}
*/
			if (blade->notch[interface].transfer) {
				// Schedule the interface for reaping at libusb's next run, coming soon
				// (The reaping occurs in razer_input_callback)
				// If we've got at least one transfer, inspire libusb to callback
				// for transfer cancels (i.e. interface reaps)
				// XXX: It looks sort of ugly, but yeah, the callback handles this
				//      part of the freeing. 
				libusb_cancel_transfer(blade->notch[interface].transfer);
				libusb_handle_events(blade->ctx);
			} else { printf("Interface %i apparently has no transfer...\n", interface); }
		}
	

		// XXX: bad juju, assuming we always have four...
		memset(blade->notch, 0, sizeof(razer_interface) * 4);
		free(blade->notch);
		blade->notch = NULL;
	}

	// Next, close the USB device
	// I assume this also performs a free(). (Hope so.)
	// XXX: I really didn't want to have to do reset, but
	//      I can't figure out how synaptics-test resets
	//      the trackpad after a use. So I just reset the
	//      whole hardware and hope for the best. :(
	libusb_reset_device(blade->handle);
	libusb_close(blade->handle);

	// XXX: Should we free the "active config descriptor", or is it shared memory
	//      that will make libusb behave badly if just pulled out from under it?
	//free(blade->cfg);
	
	// Finally, I think the last thing we can really do here is free it.
	free(blade);
}

void razer_input_interpret(unsigned char *input_buffer, size_t length) {
	uint16_t i = 0;

	if (length) {
		switch (input_buffer[0]) {
			case RAZER_MUXED_KEYBOARD:
				printf("Razer Keyboard:\n");
				for (i = 1; i < length; i++) {
					if (input_buffer[i]) {
						printf("  Keycode entered: %i\n", input_buffer[i]);
					}
				}
				break;
			case RAZER_MUXED_XTRA:
				printf("Razer Extra Buttons:\n");
				for (i = 1; i < length; i++) {
					if (input_buffer[i]) {
						if (input_buffer[i] & RAZER_IS_MACROKEY) {
							printf("  Macro key: %i\n", input_buffer[i] & 0x0F);
						} else {
							printf("  Extra key: %s\n", (input_buffer[i] == RAZER_FN_KEY ? "FN" :
							                            (input_buffer[i] == RAZER_RAZER_KEY ? "RAZER" : "UNKNOWN")
										    ) );
						}
					}
				}
				break;
			default:
				break;
		}
	}
}

void razer_input_callback(struct libusb_transfer *transfer) {
	// I'm a mother father gentleman
	// XXX: Make sure the transfer isn't just dead
	int result = 0;
	razer_interface *interface = (razer_interface*)transfer->user_data;
	uint32_t i = 0;

	if (transfer->status != LIBUSB_TRANSFER_COMPLETED) {
		switch (transfer->status) {
			case LIBUSB_TRANSFER_TIMED_OUT:
				printf("WARNING: libusb transfer timed out\n");
				break;
			case LIBUSB_TRANSFER_STALL:
				printf("ERROR: libusb transfer stalled out\n");
				return;
				break;
			case LIBUSB_TRANSFER_ERROR:
				printf("ERROR: libusb transfer error\n");
				return;
				break;
			case LIBUSB_TRANSFER_NO_DEVICE:
				printf("WARNING: Device went away (resubmitting)\n");
				libusb_submit_transfer(transfer);
				return;
				break;
			case LIBUSB_TRANSFER_CANCELLED:
				// Time to free this interface
				// (Transfer cancel for an interface means
				//  it's being reaped)
				printf("Transfer for interface %i cancelled... reaping\n", interface->interface_number);
				libusb_free_transfer(transfer);
				result = libusb_release_interface(interface->blade->handle, interface->interface_number);
				if (result) {
					printf("razer_input_callback: WARNING: Unable to release interface: %s\n", libusb_error_name(result));
				}

				if (interface->kdrv_detached) {
					result = libusb_attach_kernel_driver(interface->blade->handle, interface->interface_number);
					if (result) {
						printf("razer_input_callback: WARNING: Unable to reattach kernel driver to Blade interface %i: libusb error %s\n", interface->interface_number, libusb_error_name(result));
					} else printf("razer_input_callback: reattached kernel driver to Blade interface %i\n", interface->interface_number);
				} else { printf("kdrv never apparently detached from interface %i\n", interface->interface_number); }

				if (interface->buffer) {
					free(interface->buffer);
					interface->buffer = NULL;
					interface->buffer_size = 0;
				}

				return;
				break;
			default:
				printf("Unknown: %i\n", transfer->status);
				return;
				break;
		}
	}


	// Output all input in hex depending on what interface we got
	switch (interface->type) {
		case RAZER_KEYBOARD:
			printf("Keyboard: ");
			razer_input_interpret(interface->buffer, interface->buffer_size);
			break;
		case RAZER_MACROS:
			printf("Macro keys: ");
			razer_input_interpret(interface->buffer, interface->buffer_size);
			break;
		case RAZER_TRACKPAD:
			printf("Trackpad: ");
			break;
		case RAZER_CONTROLLER:
			// XXX: no input
			return;
			break;
		default:
			printf("UNKNOWN TYPE: interface %i: ", interface->interface_number);
			break;
	}

	for (i = 0; i < transfer->actual_length; i++) {
		printf("%02x ", transfer->buffer[i]);
	}

	printf("\n");

	interface->blade->last_input_sent = time(NULL);

	// re-submit transfer to keep the ball rolling (and also to 
	// *not* starve libusb and cause submit_transfer to block...)
	//printf("razer_input_callback: libusb submit transfer, interface %i\n", interface->interface_number);
	libusb_submit_transfer(transfer);
	//printf("razer_input_callback: libusb submit transfer, interface %i (COMPLETE)\n", interface->interface_number);
}

razer_result razer_device_set_alt(razer_device *blade, int alternate_setting, time_t timeout) {
	// We're being called to set the alternate setting on the trackpad interface
	// The strange part about this is that doing so will actually change the way
	// everything else sends data, hence this program.
	
	/* Variables */
	uint32_t i = 0;
	int result = 0;

	// Make sure the device is inited
	if (blade != NULL) {
		for (i = 0; i < blade->cfg->bNumInterfaces; i++) {
			if (blade->notch[i].type != RAZER_TRACKPAD) continue; // only do anything to the trackpad

			if (blade->cfg->interface[i].num_altsetting < 2) {
				printf("razer_device_set_alt: WARNING: this interface only has %i alt settings... not right!\n",
				       blade->cfg->interface[i].num_altsetting);
				return LIBRAZER_DEV_NOT_CORRECT;
			}

			// ok we're set
			result = libusb_set_interface_alt_setting(blade->handle, i, alternate_setting);
			if (result) {
				printf("razer_device_set_alt: WARNING: couldn't set alternate setting to trackpad: libusb error %s\n", libusb_error_name(result));
				return LIBRAZER_COULDNT_SET_ALT;
			} else {

				// We have to re-allocate the trackpad buffer because it's grown with the alt setting.
				// We also have to re-fill the transfer because the information is outdated
				// and will cause really dumb things to just start happening

				blade->notch[i].altinterface_set = alternate_setting;
				free(blade->notch[i].buffer);
				
				blade->notch[i].buffer_size = blade->cfg->interface[i].altsetting[alternate_setting].endpoint->wMaxPacketSize;
				blade->notch[i].buffer = (unsigned char*)malloc(sizeof(unsigned char*) * blade->notch[i].buffer_size);

				libusb_fill_interrupt_transfer(blade->notch[i].transfer,
				                               blade->handle,
							       blade->cfg->interface[i].altsetting[alternate_setting].endpoint->bEndpointAddress,
							       blade->notch[i].buffer,
							       blade->notch[i].buffer_size,
							       &razer_input_callback,
							       (void*)&(blade->notch[i]),
							       timeout
							      );
				return LIBRAZER_SUCCESS;
			}
		}

	}

	// Whoops. No trackpad? Weird.
	return LIBRAZER_DEV_NOT_CORRECT;
}
				
razer_result razer_device_initinput (razer_device *blade) {
	// well, if we're being called, ensure the device was inited right
	uint32_t i = 0;
	int result = 0;

	if (blade != NULL) {
		for (i = 0; i < blade->cfg->bNumInterfaces; i++) {
			if (blade->notch[i].type == RAZER_CONTROLLER) continue;

			if (!blade->notch[i].transfer) {
				printf("WARNING: Interface %i transfer request has not yet been properly created!\n", i);
				printf("WARNING: NOT submitting request for input!\n");
				continue;
			}
			// submit the interface's transfer for input
			printf("razer_device_initinput: libusb submit transfer\n");
			result = libusb_submit_transfer(blade->notch[i].transfer);
			if (result && result != LIBUSB_ERROR_BUSY) {
				printf("WARNING: could not submit transfer for interface %i: libusb error %s\n", i, libusb_error_name(result));
			}
		}
	} else {
		return LIBRAZER_DEV_NOT_INITED;
	}

	return 0;
}

int main(int argc, char **argv) {
	/* Configuration here */
	time_t seconds_to_end = 10;

	/* vars */
	razer_device *blade = NULL;
	libusb_context *ctx = NULL;
	time_t hard_time = time(NULL) + seconds_to_end;
	time_t current_time = time(NULL);
	razer_result rzi_result;
	

	/* libusb init */
	libusb_init(&ctx);

	/* find our Razer Blade */
	blade = get_razer_device(ctx, seconds_to_end);

	if (!blade) {
		printf("Something went wrong...\n");
		return 1;
	} else {
		// well.... okay then, we have a show!
		// Let's control it and run it!
		printf("All went better than expected.\n");

		// XXX: init input starts the ball rolling
		//      the callback will continue it
		rzi_result = razer_device_initinput(blade);
		if (rzi_result != LIBRAZER_SUCCESS) {
			printf("Something went wrong: error code %i\n", rzi_result);
		}
		
		rzi_result = razer_device_set_alt(blade, 1, seconds_to_end);
		if (rzi_result != LIBRAZER_SUCCESS) {
			printf("Couldn't set alt interface: error %i\n", rzi_result);
		}

		// Event loop/program duration stuff...
		for (current_time; 
		     current_time < hard_time;
		     current_time = time(NULL)) {
		     	// Make libusb handle events. We do no other input;
			// it is OK to block on libusb's event handler.
//			printf("libusb handle events\n");
			libusb_handle_events(ctx);
		}
		free_razer_device(blade);
		blade = NULL;
	}

	libusb_exit(ctx);

	/* guh */
	return 0;
}

/*
 * [1]: So, currently, Razer Blade sets its input interfaces to types as such:
 * 	0 - Trackpad
 * 	1 - Keyboard
 * 	2 - Macro-Keyboard
 * 	3 - Controller (bulk transfer interface probably for controlling mouse)
 *
 * I set up the enum to work with this quickly, i.e. the interface indicates its
 * type. This isn't future-proof (future firmware updates could break this, future
 * hardware may not conform to it), but it should suffice for now.
 */
